// Copyright (c) 2007 Brian Harry
package org.jythonx.utils;

import org.python.compiler.*;
import org.python.core.Py;
import org.python.core.PyBuiltinFunction;
import org.python.core.PyBuiltinFunctionNarrow;
import org.python.core.PyBuiltinFunctionWide;
import org.python.core.PyObject;
import org.python.core.PyMethodDescr;
import java.lang.reflect.*;
import java.io.*;

/**
 * Utility class that generates boilerplate classes/instances at runtime.
 */
public final class Boilerplate{

    static final String dumpDirectory = System.getProperty("boilerplateDumpDirectory", "boilerplate");
    static final String fileDump = System.getProperty("boilerplateFileDump", null);
    static final String loadFromClassLoader = System.getProperty("boilerplateLoadClass", null);
    static final String verboseExceptionDump = System.getProperty("boilerplateExceptionPrint", null);

    //used to define the class from the generated bytes
    static BoilerplateClassLoader loader = new BoilerplateClassLoader(PyObject.class.getClassLoader());

    static class BoilerplateClassLoader extends ClassLoader{
	BoilerplateClassLoader(ClassLoader cl){ super(cl); }
        public Class getClass(String name, byte[] data){
            return defineClass(name, data, 0, data.length);
	}
    }

    /** Hooks that are invoked before and after __call__ and inst_call code bodies are generated */
    public static interface PrePostHooks{
        /** Called before any code has been produced in __call__ or inst_call 
         * If true is returned, the main body of the code is not produced.
         */
	boolean pre(Code code) throws IOException;
        /** Called after invokevirtual or invokestatic has been invoked in __call__ or inst_call */
        void post(Code code) throws IOException;
    }

    /**
     * Adds this code afterwards:
     * if(ret == null)
     *     return Py.NotImplemented
     */
    public static PrePostHooks NotImplementedCheck = new PrePostHooks(){
	    public boolean pre(Code code) throws IOException{ return false; }
            public void post(Code code) throws IOException{
		code.dup();
                Label l = code.getLabel();
                code.ifnonnull(l);
                int fref = code.pool.Fieldref("org/python/core/Py", "NotImplemented", "Lorg/python/core/PyObject;");
                code.getstatic(fref);
                code.areturn();
                l.setPosition();
            }
	};

    public static PrePostHooks PyStringConvert = new PrePostHooks(){
            public boolean pre(Code code) throws IOException{
		int pstring = code.pool.Class("org/python/core/PyString");
                code.new_(pstring);
                code.dup();
                return false;
            }
	    public void post(Code code) throws IOException{
                int toString = code.pool.Methodref("java/lang/String", "toString", "()Ljava/lang/String;");
                code.invokevirtual(toString);
		int init = code.pool.Methodref("org/python/core/PyString", "<init>", "(Ljava/lang/String;)V");
                code.invokespecial(init);
            }
	};

    public static PrePostHooks PyIntegerConvert1 = new PrePostHooks(){
            public boolean pre(Code code) throws IOException{
		int pstring = code.pool.Class("org/python/core/PyInteger");
                code.new_(pstring);
                code.dup();
                return false;
            }
	    public void post(Code code) throws IOException{
		int init = code.pool.Methodref("org/python/core/PyInteger", "<init>", "(I)V");
                code.invokespecial(init);
            }
	};

    public static PrePostHooks PyIntegerConvert2 = new PrePostHooks(){
            public boolean pre(Code code) throws IOException{
		int pstring = code.pool.Class("org/python/core/PyInteger");
                code.new_(pstring);
                code.dup();
                return false;
            }
	    public void post(Code code) throws IOException{
                int toInteger = code.pool.Methodref("java/lang/Integer", "valueOf", "()Ljava/lang/Integer;");
                code.invokevirtual(toInteger);
		int init = code.pool.Methodref("org/python/core/PyInteger", "<init>", "(Ljava/lang/Integer;)V");
                code.invokespecial(init);
            }
	};

    private Boilerplate(){}



    public static void fillDictWithNarrows(PyObject dict, Class clazz, Object[][] buildData){
        try{
            for(int i = 0; i < buildData.length; i += 2){
                Object[] data = buildData[i];
                PyBuiltinFunction function = Boilerplate.newNarrow(clazz, 
                                                                  (String)data[0], 
                                                                  (int[])data[1], 
                                                                  (boolean[])data[2],
								  (Boilerplate.PrePostHooks)data[3]);
                Object[] methodDescr = buildData[i + 1];
                String name = (String)methodDescr[0];
                int min = ((int[])methodDescr[1])[0];
                int max = ((int[])methodDescr[1])[1];
                dict.__setitem__(name, new PyMethodDescr(name, clazz, min, max, function));
	    }
        } catch(Throwable t){
            throw Py.RuntimeError(t.getMessage());
        }
    }

    public static void fillDictWithWides(PyObject dict, Class clazz, Object[][] buildData){
        try{
            for(int i = 0; i < buildData.length; i += 2){
                Object[] data = buildData[i];
                PyBuiltinFunction function = Boilerplate.newWide(clazz, 
                                                                  (String)data[0], 
                                                                  data[1] == Boolean.TRUE, 
                                                                  data[2] == Boolean.TRUE,
								  (Boilerplate.PrePostHooks)data[3]);
                Object[] methodDescr = buildData[i + 1];
                String name = (String)methodDescr[0];
                int min = ((int[])methodDescr[1])[0];
                int max = ((int[])methodDescr[1])[1];
                dict.__setitem__(name, new PyMethodDescr(name, clazz, min, max, function));
	    }
        } catch(Throwable t){
            t.printStackTrace();
            throw Py.RuntimeError(t.getMessage());
        }
    }




    /**
     * Produces at runtime a PyBuiltinFunctionNarrow instance that targets a set of virtual or static methods of the
     * passed in class object.  The narrows and statics parameters reflect each other.  narrows holds the number of
     * PyObjects that will be passed in while statics determines if invokevirtual or invokestatic will be used to
     * invoke the method.
     * 
     * This produces a class of this form:
     * newNarrow(PyBoolean.class, "toString", new int[]{0}, new boolean[]{true}, null);
     * package org.python.core;
     * public class PyBooleantoString extends PyBuiltinFunctionNarrow{
     *
     *    public PyBoolean self
     *    public PyObject getSelf(){ return self; }
     *    public PyBooleantoString(PyObject self, PyBuiltinFunctionNarrow.Info info){
     *        super(info);
     *        this.self = self;
     *    }
     *
     *    public PyBooleantoString(){
     *        super(null);
     *    }
     *
     *    public PyBuiltinFunction makeBound(PyObject self){
     *        return new PyBooleantoString(self, info);
     *    }
     *
     *    public PyObject __call__(){
     *        return PyBoolean.toString(self);
     *    }
     *
     *    public PyObject inst_call(PyObject self){
     *        return PyBoolean.toString(self);
     *    }
     * }
     * Here we see that PyBoolean is the target class.  The method(s) that will be invoked is called toString.
     * From looking at the narrows param, there is going to be one method called.  This particular method will
     * take 0 parameters and is a static method.  A invocation like this:
     * newNarrow(PyBoolean.class, "__and__", new int[]{1}, new boolean[]{false}, null);
     * means that a instance method called "__and__" will be targeted and it will take one parameter.
     *
     * If static, the first argument to the static method will be the targeted PyObject.  Arguments in this case
     * come after the first parameter.  When virtual methods, the parameters always start at argument 0.
     *
     * @param pyobject the targeted class
     * @param methodName the method name
     * @param narrows the number of arguments for the method at the index
     * @param statics if the method at the index is static or virtual, 
                      null is equivilent to an array filled with false
     * @param ppi are hooks that are invoked prior and post the invocation of method.  Allows greater customisation.
     * @throws Exception if there was a problem generating the function
     */
    public static PyBuiltinFunctionNarrow newNarrow(Class pyobject, String methodName, int[] narrows, boolean[] statics, PrePostHooks pphks) throws Exception{

        String alternativeReturnType = null;        
        if(methodName.split(":").length != 1){
            String[] pieces = methodName.split(":");
	    alternativeReturnType = pieces[pieces.length - 1];
            methodName = methodName.substring(0, methodName.length() - (alternativeReturnType.length() + 1));
        }
        String name = pyobject.getName();
        String newName = boilerplateName(pyobject, methodName);
        PyBuiltinFunctionNarrow rv = (PyBuiltinFunctionNarrow)testLoad(newName);
        if(rv != null) return rv;

        ClassFile cf = new ClassFile(newName, "org/python/core/PyBuiltinFunctionNarrow", ClassFile.PUBLIC);
        String newNameTranslated = newName.replace(".", "/");
        String pyClassNameTranslated = name.replace(".", "/");        

        cf.addField("self", "L" + pyClassNameTranslated + ";", ClassFile.PUBLIC);
        int selfRef = cf.pool.Fieldref(newNameTranslated, "self",  "L" + pyClassNameTranslated + ";");
        int classRef = cf.pool.Class(pyClassNameTranslated);
        int ncRef = cf.pool.Class(newNameTranslated);

        addGetSelf(cf, selfRef);
        addInits(cf, classRef, selfRef, true);
        addMakeBound(cf, ncRef, newNameTranslated);
        

        for(int i = 0; i < narrows.length; i ++){
            int args = narrows[i];            
            boolean isStatic = statics != null? statics[i]: false;
            StringBuilder signature = new StringBuilder();
            StringBuilder signature2 = new StringBuilder();
            StringBuilder signature3 = new StringBuilder();
            StringBuilder methodSignature = new StringBuilder();
            signature2.append("(L").append(pyClassNameTranslated).append(';');
            signature3.append("(Lorg/python/core/PyObject;");
            if(args == -2){
                signature.append("([Lorg/python/core/PyObject;[Ljava/lang/String;");
                signature2.append("[Lorg/python/core/PyObject;[Ljava/lang/String;");
		signature3.append("[Lorg/python/core/PyObject;[Ljava/lang/String;");
                methodSignature.append("([Lorg/python/core/PyObject;[Ljava/lang/String;");
                System.out.println(signature);
                System.out.println(signature2);
                System.out.println(signature3);
                System.out.println(methodSignature);
            }
            else{
                signature.append("(");
                methodSignature.append("(");
                if(isStatic)
                    methodSignature.append("L").append(pyClassNameTranslated).append(";");

                for(int j = 0; j < args; j ++){
                    signature.append("Lorg/python/core/PyObject;");
                    signature2.append("Lorg/python/core/PyObject;");
                    signature3.append("Lorg/python/core/PyObject;");
                    methodSignature.append("Lorg/python/core/PyObject;");
                }
	    }
            signature.append(")Lorg/python/core/PyObject;");
            signature2.append(")Lorg/python/core/PyObject;");
            signature3.append(")Lorg/python/core/PyObject;");
            methodSignature.append(alternativeReturnType != null? ")" + alternativeReturnType: ")Lorg/python/core/PyObject;");

            Code __call__ = cf.addMethod("__call__", signature.toString(), ClassFile.PUBLIC);
            int mref = cf.pool.Methodref(pyClassNameTranslated, methodName, methodSignature.toString());
            boolean replaced = false;
            if(pphks != null)
		replaced = pphks.pre(__call__);
            if(!replaced){
                __call__.aload(0);
                __call__.getfield(selfRef);
                if(args == -2){
                    __call__.aload(1);
                    __call__.aload(2);
		}
                else{
                    for(int j = 0; j < args; j ++)
                        __call__.aload(j + 1);
                }
                if(isStatic)
                    __call__.invokestatic(mref);
                else
                    __call__.invokevirtual(mref);
            }
            if(pphks != null)
                pphks.post(__call__);
            if(alternativeReturnType != null && alternativeReturnType.equals("V")){
		int PyNone = __call__.pool.Fieldref("org/python/core/Py", "None", "Lorg/python/core/PyObject;");
                __call__.getstatic(PyNone);
            }
            __call__.areturn();

            Code inst_call = cf.addMethod("inst_call", signature3.toString(), ClassFile.PUBLIC);
            replaced = false;
            if(pphks != null)
                replaced = pphks.pre(inst_call);
            if(!replaced){
                inst_call.aload(1);
                inst_call.checkcast(classRef);
                if(args == -2){
                    inst_call.aload(2);
                    inst_call.aload(3);
                }
                else{
                    for(int j = 0; j < args; j ++)
                        inst_call.aload(j + 2);
		}
                if(isStatic)
                    inst_call.invokestatic(mref);
                else
                    inst_call.invokevirtual(mref);
            }
            if(pphks != null)
                pphks.post(inst_call);
            if(alternativeReturnType != null && alternativeReturnType.equals("V")){
		int PyNone = inst_call.pool.Fieldref("org/python/core/Py", "None", "Lorg/python/core/PyObject;");
                inst_call.getstatic(PyNone);
            }
	    inst_call.areturn();
        }

        return (PyBuiltinFunctionNarrow)writeAndInstantiateClass(cf, newName);
    }

    public static PyBuiltinFunctionWide newWide(Class pyobject, String methodName, boolean takesKeywords, boolean isStatic, PrePostHooks pphks) throws Exception{

        String alternativeReturnType = null;        
        if(methodName.split(":").length != 1){
            String[] pieces = methodName.split(":");
	    alternativeReturnType = pieces[pieces.length - 1];
            methodName = methodName.substring(0, methodName.length() - (alternativeReturnType.length() + 1));
        }
        String name = pyobject.getName();
        String newName = boilerplateName(pyobject, methodName);
        PyBuiltinFunctionWide rv = (PyBuiltinFunctionWide)testLoad(newName);
        if(rv != null) return rv;

        ClassFile cf = new ClassFile(newName, "org/python/core/PyBuiltinFunctionWide", ClassFile.PUBLIC);
        String newNameTranslated = newName.replace(".", "/");
        String pyClassNameTranslated = name.replace(".", "/");        

        cf.addField("self", "L" + pyClassNameTranslated + ";", ClassFile.PUBLIC);
        int selfRef = cf.pool.Fieldref(newNameTranslated, "self",  "L" + pyClassNameTranslated + ";");
        int classRef = cf.pool.Class(pyClassNameTranslated);
        int ncRef = cf.pool.Class(newNameTranslated);

        addGetSelf(cf, selfRef);
        addInits(cf, classRef, selfRef, false);
        addMakeBound(cf, ncRef, newNameTranslated);
        

        StringBuilder methodSignature = new StringBuilder();
        StringBuilder methodSignature2 = new StringBuilder();
        methodSignature.append("([Lorg/python/core/PyObject;");
        methodSignature2.append("([Lorg/python/core/PyObject;[Ljava/lang/String;");
        methodSignature.append(alternativeReturnType != null? ")" + alternativeReturnType: ")Lorg/python/core/PyObject;");
        methodSignature2.append(alternativeReturnType != null? ")" + alternativeReturnType: ")Lorg/python/core/PyObject;");


        String ic1_sig = "(Lorg/python/core/PyObject;[Lorg/python/core/PyObject;)Lorg/python/core/PyObject;";
	Code inst_call1 = cf.addMethod("inst_call", ic1_sig, ClassFile.PUBLIC);
        boolean replaced = false;
        if(pphks != null)
	    replaced = pphks.pre(inst_call1);
        if(!replaced){
             inst_call1.aload(1);
             inst_call1.checkcast(classRef);
             inst_call1.aload(2);
             int mref;
             if(!takesKeywords)
                 mref = cf.pool.Methodref(pyClassNameTranslated, methodName, methodSignature.toString());
             else{
                 mref = cf.pool.Methodref(pyClassNameTranslated, methodName, methodSignature2.toString());
	        int NoKeywords = inst_call1.pool.Fieldref("org/python/core/Py", "NoKeywords", "[Ljava/lang/String;");
                inst_call1.getstatic(NoKeywords);
             }
             if(isStatic)
                 inst_call1.invokestatic(mref);
             else
                 inst_call1.invokevirtual(mref);
        }
        if(pphks != null)
            pphks.post(inst_call1);
        if(alternativeReturnType != null && alternativeReturnType.equals("V")){
	    int PyNone = inst_call1.pool.Fieldref("org/python/core/Py", "None", "Lorg/python/core/PyObject;");
            inst_call1.getstatic(PyNone);
        }
        inst_call1.areturn();

        if(takesKeywords){
            String ic2_sig = "(Lorg/python/core/PyObject;[Lorg/python/core/PyObject;[Ljava/lang/String;)Lorg/python/core/PyObject;";
           Code inst_call2 = cf.addMethod("inst_call", ic2_sig, ClassFile.PUBLIC);
           replaced = false;
           if(pphks != null)
	        replaced = pphks.pre(inst_call2);
           if(!replaced){
                 inst_call2.aload(1);
                 inst_call2.checkcast(classRef);
                 inst_call2.aload(2);
                 inst_call2.aload(3);
                 int mref = cf.pool.Methodref(pyClassNameTranslated, methodName, methodSignature2.toString());
                 if(isStatic)
                     inst_call2.invokestatic(mref);
                 else
                     inst_call2.invokevirtual(mref);
           }
           if(pphks != null)
               pphks.post(inst_call2);
            if(alternativeReturnType != null && alternativeReturnType.equals("V")){
	        int PyNone = inst_call2.pool.Fieldref("org/python/core/Py", "None", "Lorg/python/core/PyObject;");
                inst_call2.getstatic(PyNone);
            }
            inst_call2.areturn();
	}
        return (PyBuiltinFunctionWide)writeAndInstantiateClass(cf, newName);
    }




    /**
     * Simple utility method that produces a PyBuiltinFunction that targets a static method of a Class.
     *
     * @param clazz the Class
     * @param name the name of the method
     * @return the builtin function
     * @throws Exception if there is a problem generating the function
     */
    public static PyBuiltinFunction newStaticBuiltinFunction(Class clazz, String methodName, Class[] params) throws Exception{
        String newName = boilerplateName(clazz, methodName);
        PyBuiltinFunction rv = (PyBuiltinFunction)testLoad(newName);
        if(rv != null) return rv;

        Method m = clazz.getMethod(methodName, params);
        Class returnType = m.getReturnType();
        if(!Modifier.isStatic(m.getModifiers())){
            throw new IllegalArgumentException("non static method passed into staticBuiltinFunction");
	}

        ClassFile cf = new ClassFile(newName, "org/python/core/PyBuiltinFunction", ClassFile.PUBLIC);

        Code init = cf.addMethod("<init>", "()V", ClassFile.PUBLIC);        
        init.aload(0);
        int getInfo = cf.pool.Methodref("org/jythonx/utils/Boilerplate", "getInfo", "(Ljava/lang/String;I)Lorg/python/core/PyBuiltinFunction$Info;");
        init.ldc(methodName);
        init.iconst(params.length);
        init.invokestatic(getInfo);
        int sref = cf.pool.Methodref("org/python/core/PyBuiltinFunction", "<init>", "(Lorg/python/core/PyBuiltinFunction$Info;)V");
        init.invokespecial(sref);
        init.return_();

        Code __call__ = cf.addMethod("__call__", "([Lorg/python/core/PyObject;[Ljava/lang/String;)Lorg/python/core/PyObject;", ClassFile.PUBLIC);
        int infoRef = cf.pool.Fieldref("org/python/core/PyBuiltinFunction", "info", "Lorg/python/core/PyBuiltinFunction$Info;");
        __call__.aload(0);
        __call__.getfield(infoRef);
        __call__.aload(1);
        __call__.aload(2);
        int testInvocation = cf.pool.Methodref("org/jythonx/utils/Boilerplate", "testInvocation", "(Lorg/python/core/PyBuiltinFunction$Info;[Lorg/python/core/PyObject;[Ljava/lang/String;)V");
        __call__.invokestatic(testInvocation);
        for(int i = 0; i < params.length; i ++){
            __call__.aload(1);
            if(params[i].isArray()){
                if(i == 0){
                    continue;
                }
                __call__.iconst(i);
                __call__.iconst(params.length - i);
                int pyObj = __call__.pool.Class("org/python/core/PyObject");
                __call__.anewarray(pyObj);
                int pyObjArr = __call__.getLocal("pyObjArr");
                __call__.astore(pyObjArr);
                __call__.aload(pyObjArr);
                __call__.iconst(0);
                __call__.iconst(params.length - i);
                int icopy = __call__.pool.Methodref("java/lang/System", "arraycopy", "(Ljava/lang/Object;ILjava/lang/Object;II)V");
                __call__.invokestatic(icopy);
                __call__.aload(pyObjArr);
                __call__.freeLocal(pyObjArr);
	    } else{
                __call__.iconst(i);
                __call__.aaload();
            }
        }
        String cname = CodeUtils.translateNonPrimitive(clazz);
        String signature = CodeUtils.createSignature(params, returnType);
        int smref = __call__.pool.Methodref(cname, methodName, signature);
	__call__.invokestatic(smref);
        if(returnType.isPrimitive()){
            CodeUtils.boxPrimitive(returnType, __call__, false);
            if(returnType == Void.TYPE){
                int none = __call__.pool.Fieldref("org/python/core/Py", "None", "Lorg/python/core/PyObject;");
                __call__.getstatic(none);
                __call__.areturn();
            }
        }
        if(PyObject.class.isAssignableFrom(returnType)){
	    __call__.areturn();
	}
	else{
            int java2py = __call__.pool.Methodref("org/python/core/Py", "java2py", "(Ljava/lang/Object;)Lorg/python/core/PyObject;");
            __call__.invokestatic(java2py);
            __call__.areturn();
	}
        return (PyBuiltinFunction)writeAndInstantiateClass(cf, newName);
    }

    static void addMakeBound(ClassFile cf, int ncRef, String newNameTranslated) throws IOException{
        Code makeBound = cf.addMethod("makeBound", "(Lorg/python/core/PyObject;)Lorg/python/core/PyBuiltinFunction;", ClassFile.PUBLIC);
        makeBound.new_(ncRef);
        makeBound.dup();
        makeBound.aload(1);
        makeBound.aload(0);
        int infoField = cf.pool.Fieldref("org/python/core/PyBuiltinFunction", "info", "Lorg/python/core/PyBuiltinFunction$Info;");
        makeBound.getfield(infoField);
        int superInit2 = cf.pool.Methodref(newNameTranslated, "<init>", "(Lorg/python/core/PyObject;Lorg/python/core/PyBuiltinFunction$Info;)V");
        makeBound.invokespecial(superInit2);
        makeBound.areturn();
    }

    static void addInits(ClassFile cf, int classRef, int selfRef, boolean narrow) throws IOException{
        Code init = cf.addMethod("<init>", "(Lorg/python/core/PyObject;Lorg/python/core/PyBuiltinFunction$Info;)V", ClassFile.PUBLIC);
        init.aload(0);
        init.aload(2);
        int superInit = cf.pool.Methodref("org/python/core/PyBuiltinFunction" + (narrow? "Narrow" : "Wide"), "<init>", "(Lorg/python/core/PyBuiltinFunction$Info;)V");
        init.invokespecial(superInit);
        init.aload(0);
        init.aload(1);
        init.checkcast(classRef);
        init.putfield(selfRef);
        init.return_();

        Code init2 = cf.addMethod("<init>", "()V", ClassFile.PUBLIC);
        init2.aload(0);
        init2.aconst_null();
        init2.invokespecial(superInit);
        init2.return_();
    }

    static void addGetSelf(ClassFile cf, int selfRef) throws IOException{
        Code getSelf = cf.addMethod("getSelf", "()Lorg/python/core/PyObject;", ClassFile.PUBLIC);
        getSelf.aload(0);
        getSelf.getfield(selfRef);
        getSelf.areturn();
    }

    public static void testInvocation(PyBuiltinFunction.Info info, PyObject[] args, String[] kwords){
        int argl = args.length;
        if(kwords.length != 0)
	    throw info.unexpectedCall(argl, true);
        int min = info.getMinargs();
        if(argl < min)
            throw info.unexpectedCall(argl, false);
        int max = info.getMaxargs();
        if(max < argl)
            throw info.unexpectedCall(argl, false);        
    }

    public static PyBuiltinFunction.Info getInfo(String name, int args){
        return new PyBuiltinFunction.DefaultInfo(name, args);
    }

    static Object writeAndInstantiateClass(ClassFile cf, String newName) throws Exception{
        String[] tokens = newName.split("\\.");
        ByteArrayOutputStream baos = new ByteArrayOutputStream(1000);
        cf.write(baos);
        byte[] data = baos.toByteArray();
        if(fileDump != null){
            File f = new File(dumpDirectory, tokens[tokens.length - 1] + ".class");
	    FileOutputStream fos = new FileOutputStream(f);
            fos.write(data, 0, data.length);
        }
        Class x = loader.getClass(newName, data);        
        return x.newInstance();
    }

    static Object testLoad(String newName){
        if(loadFromClassLoader != null){
	    try{
                Class c = loader.loadClass(newName);
                return c.newInstance();
            } catch(ClassNotFoundException e){
		if(verboseExceptionDump != null)
                    e.printStackTrace();
            } catch(InstantiationException e){
                if(verboseExceptionDump != null)
		    e.printStackTrace();
            } catch(IllegalAccessException e){
                if(verboseExceptionDump != null)
                    e.printStackTrace();
            }
        }
        return null;
    }

    static String boilerplateName(Class c, String methodName){
        String name = c.getName();
        String[] tokens = name.split("\\.");
        return "org.python.boilerplate." + tokens[tokens.length - 1] + methodName;
    }

}