// Copyright (c) 2007 Brian Harry
package org.jythonx.utils;

import org.python.compiler.*;
import java.io.IOException;

/**
 * Utility methods for working with Code objects.
 */
public final class CodeUtils{

    private final static String vOf = "valueOf";
    private final static String boolString = "java/lang/Boolean";
    private final static String byteString = "java/lang/Byte";
    private final static String charString = "java/lang/Character";
    private final static String doubString = "java/lang/Double";
    private final static String floatString = "java/lang/Float";
    private final static String intString = "java/lang/Integer";
    private final static String longString = "java/lang/Long";
    private final static String shortString = "java/lang/Short";

    private CodeUtils(){}

    /** Boxes a primitive up */
    public static void boxPrimitive(Class castClass, Code c, boolean return_) throws IOException{
        int ref;
        if(castClass == Boolean.TYPE) 
            ref = c.pool.Methodref(boolString, vOf, "(Z)Ljava/lang/Boolean;");        
        else if(castClass == Byte.TYPE)
            ref = c.pool.Methodref(byteString, vOf, "(B)Ljava/lang/Byte;");
        else if(castClass == Character.TYPE)
            ref = c.pool.Methodref(charString, vOf, "(C)Ljava/lang/Character;");
        else if(castClass == Double.TYPE)
            ref = c.pool.Methodref(doubString, vOf, "(D)Ljava/lang/Double;");
        else if(castClass == Float.TYPE)
            ref = c.pool.Methodref(floatString, vOf, "(F)Ljava/lang/Float;");
        else if(castClass == Integer.TYPE)
            ref = c.pool.Methodref(intString, vOf, "(I)Ljava/lang/Integer;");
        else if(castClass == Long.TYPE)
            ref = c.pool.Methodref(longString, vOf, "(J)Ljava/lang/Long;");
        else if(castClass == Short.TYPE) 
            ref = c.pool.Methodref(shortString, vOf, "(S)Ljava/lang/Short;");
	else 
            throw new IllegalArgumentException("Class " + castClass + " is not a primitive");
        c.invokestatic(ref);
	if(return_) c.areturn();
    }
    
    /** Turns a boxed primitive into a primitive, optional return call */
    public static void toPrimitive(Class castClass, Code c, boolean return_) throws IOException{
        int ref = -1;
        if(castClass == Boolean.TYPE)
	    ref = c.pool.Methodref(boolString, "booleanValue", "()Z");
        else if(castClass == Byte.TYPE)
            ref = c.pool.Methodref(byteString, "byteValue", "()B");
        else if(castClass == Character.TYPE)
            ref = c.pool.Methodref(charString, "charValue", "()C");
        else if(castClass == Double.TYPE)
            ref = c.pool.Methodref(doubString, "doubleValue", "()D");
        else if(castClass == Float.TYPE)
            ref = c.pool.Methodref(floatString, "floatValue", "()F");
        else if(castClass == Integer.TYPE)
            ref = c.pool.Methodref(intString, "intValue", "()I");
        else if(castClass == Long.TYPE)
            ref = c.pool.Methodref(longString, "longValue", "()J");
        else if(castClass == Short.TYPE)
            ref = c.pool.Methodref(shortString, "shortValue", "()S");
        else if(castClass == Void.TYPE);
	else 
            throw new IllegalArgumentException("Class " + castClass + " is not a primitive");
        if(castClass != Void.TYPE)
	    c.invokevirtual(ref);
        if(return_){
            returnHelper(castClass, c);
        }
    }

    /** Turns name.of.something into Lname/of/something; */
    public static String translate(Class c){
        if(c.isPrimitive()){
	    if(c == Boolean.TYPE) return "Z";
            else if(c == Byte.TYPE) return "B";
            else if(c == Character.TYPE) return "C";
            else if(c == Double.TYPE) return "D";
            else if(c == Float.TYPE) return "F";
            else if(c == Integer.TYPE) return "I";
            else if(c == Long.TYPE) return "J";
            else if(c == Short.TYPE) return "S";
            else if(c == Void.TYPE) return "V";
        }
        String name = c.getName();
        name = name.replace(".", "/");
        if(c.isArray()) return name;
        else
            return "L" + name + ";";
    }

    /** Turns name.of.something into name/of/something */
    public static String translateNonPrimitive(Class c){
	String name = c.getName();
        name = name.replace(".", "/");
        return name;
    }

    /** Builds a signature String */
    public static String createSignature(Class[] params, Class returnValue){
	StringBuilder sb = new StringBuilder();
        sb.append("(");
        for(Class param : params)
            sb.append(translate(param));
        sb.append(")");
        sb.append(translate(returnValue));
        return sb.toString();
    }

    /** Casts to the class */
    public static void castTo(Class castTo, Code code) throws IOException{
        int cref = -1;
        if(castTo.isPrimitive()){
	    if(castTo == Boolean.TYPE)
                cref = code.pool.Class(boolString);
            else if(castTo == Byte.TYPE)
                cref = code.pool.Class(byteString);
            else if(castTo == Character.TYPE)
                cref = code.pool.Class(charString);
            else if(castTo == Double.TYPE)
                cref = code.pool.Class(doubString);
            else if(castTo == Float.TYPE)
                cref = code.pool.Class(floatString);
            else if(castTo == Integer.TYPE)
                cref = code.pool.Class(intString);
            else if(castTo == Long.TYPE)
                cref = code.pool.Class(longString);
            else if(castTo == Short.TYPE)
                cref = code.pool.Class(shortString);
        } 
        else cref = code.pool.Class(translateNonPrimitive(castTo));
        code.checkcast(cref);
    }

    /** Calls appropiate return for passed in class */
    public static void returnHelper(Class returnClass, Code c) throws IOException{
        if(returnClass == Void.TYPE) c.return_();
        else if(returnClass.isPrimitive()){
	    if(returnClass == Boolean.TYPE) c.ireturn();
            else if(returnClass == Byte.TYPE) c.ireturn();
            else if(returnClass == Character.TYPE) c.ireturn();
            else if(returnClass == Double.TYPE) c.dreturn();
            else if(returnClass == Float.TYPE) c.freturn();
            else if(returnClass == Integer.TYPE) c.ireturn();
            else if(returnClass == Long.TYPE) c.lreturn();
            else if(returnClass == Short.TYPE) c.ireturn();
	}
        else c.areturn();
    }

    /** puts the locals on the stack */
    public static void loadLocals(Class[] cparams, Code c) throws IOException{
        for(int i = 0, j = 1; i < cparams.length; i ++){             
            Class pc = cparams[i];            
            if(pc.isPrimitive()){
	        if(pc == Boolean.TYPE){
                    c.iload(j);
                    j++;
                }
                else if(pc == Byte.TYPE){ 
                    c.iload(j);
                    j++;
                }
                else if(pc == Character.TYPE){
                    c.iload(j);
                    j++;
                }
                else if(pc == Double.TYPE){
                    c.dload(j);
                    j += 2;
                }
                else if(pc == Float.TYPE){
                     c.fload(j);
                     j ++;
		}
                else if(pc == Integer.TYPE){
                    c.iload(j);
                    j++;
                }
                else if(pc == Long.TYPE){
                      c.lload(j);
                      j += 2;
                }
                else if(pc == Short.TYPE){
                     c.iload(j);
                     j ++;
		}
            } else{
                  c.aload(j);
                  j ++;
            }
        }
    }

    /** Transforms locals to a PyObject[], after exiting the arrays will be on the stack */
    public static void localsToPyObjectArray(Class[] cparams, Code c) throws IOException{
        int pyref = c.pool.Methodref("org/python/core/Py", "java2py", "(Ljava/lang/Object;)Lorg/python/core/PyObject;"); 
        c.iconst(cparams.length); 
        int pyClassRef = c.pool.Class("org/python/core/PyObject"); 
        c.anewarray(pyClassRef);
        for(int i = 0, j = 1; i < cparams.length; i++){ 
            c.dup();
            Class pc = cparams[i]; 
            c.iconst(i); 
            if(pc.isPrimitive()){ 
	        if(pc == Boolean.TYPE){ 
                    c.iload(j); 
                    j++; 
                } 
                else if(pc == Byte.TYPE){ c.iload(j); 
                    j++; 
                } 
                else if(pc == Character.TYPE){ 
                    c.iload(j);
                    j++;
                }
                else if(pc == Double.TYPE){
                    c.dload(j);
                    j += 2;
                }
                else if(pc == Float.TYPE){
                     c.fload(j);
                     j ++;
		}
                else if(pc == Integer.TYPE){
                    c.iload(j);
                    j++;
                }
                else if(pc == Long.TYPE){
                      c.lload(j);
                      j += 2;
                }
                else if(pc == Short.TYPE){
                     c.iload(j);
                     j++;
		}
                CodeUtils.boxPrimitive(pc, c, false);
            } else{
                  c.aload(j);
                  j ++;
            }
            c.invokestatic(pyref); 
            c.aastore();
        }
    }

    /** loads a class */
    public static void loadClass(Class c, Code code) throws IOException{
        String name = translateNonPrimitive(c);
        int clazz = code.pool.Class(name);
        code.ldc(clazz);
    }

    /** Convenience print function, Code lacking this */
    public static void printObject(Code code) throws IOException{
        int localObj = code.getLocal("localObj");
        code.astore(localObj);
        code.getstatic("java/lang/System", "out", "Ljava/io/PrintStream;");
        code.aload(localObj);
        code.invokevirtual("java/io/PrintStream", "println", "(Ljava/lang/Object;)V");
        code.aconst_null();
        code.astore(localObj);
        code.freeLocal(localObj);
    }

}